/*->c.pcx */


#include "stdafx.h"

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <locale.h>

#include "os.h"
#include "wimp.h"
#include "wimpt.h"
#include "werr.h"
#include "wos.h"
#include "flex.h"
#include "transform.h"
#include "process.h"


#include "err.h"
#include "fsx.h"
#include "task.h"
#include "alloc.h"
#include "trans.h"
#include "etc.h"
#include "bf.h"


#include "str.h"

#include "main.h"
#include "im.h"

#include "pcx.h"




/*****************************************************************************/

/*
**      PCX File Header Structure
*/


static int tdbl;
static int dbl;
static int pcxplanes=1; /* use seemingly correct interlaced planes for 4bpp */

static int pcxxres=96;
static int pcxyres=96;
static int txres;
static int tyres;


typedef char XBYTE;



typedef struct pcxhdr
{
 XBYTE    manufacturer;  /*      10 == ZSoft */
 XBYTE    version;       /*      0 = Paintbrush 2.5
                             ** 2 = 2.8 with palette
                             ** 3 = 2.8 without palette
                             ** 4 = PC Paintbrush for Windows
                             ** 5 = Version 3.0+ and Publishers Paintbrush */

 XBYTE   encode;                 /*      1 = PCX RLE */
 XBYTE   bitsPerPixel;           /* 1, 2, 4, or 8 */
 short   minX;                                   /* Window dimensions */
 short   minY;
 short   maxX;
 short   maxY;
 short   horzDpi;                /* Horizontal dots per inch */
 short   vertDpi;                /* Vertical dots per inch */
 XBYTE   pal1[48];               /* palette */
 XBYTE   junk;
 XBYTE   planes;                 /* number of color planes */
 short   bytesPerLine;           /* number of bytes per scanline */
 short   paltype;                /* 1 = Color, 2 = GrayScale */
 short   hScreenSize;            /* Horizontal Screen Size */
 short   vScreenSize;            /* Vertical Screen Size */
 XBYTE   filler[54];

} PCXHDR;

/*
**      Note: Version 5 images may contain a 256 color palette at the end of the
**      file.  Check for it by checking if the byte 769 bytes before the end of
**      the file is 12.
*/


#define PCXPALSIZE 769

typedef struct pcxpal256
{
 XBYTE         pal2[PCXPALSIZE];
} PCXPAL256;


/*
**      Bit masks for planar decoding
*/
static XBYTE LowMasks[4]  ={ 0x01, 0x02, 0x04, 0x08 };
static XBYTE HighMasks[4] ={ 0x10, 0x20, 0x40, 0x80 };

/*
**      Default 16 color VGA palette
*/

static char DefaultPalette[48]=
{
           0,   0,   0,
           0,   0, 255,
           0, 255,   0,
           0, 255, 255,
         255,   0,   0,
         255,   0, 255,
         255, 255,   0,
         255, 255, 255,
          85,  85, 255,
          85,  85,  85,
           0, 170,   0,
         170,   0,   0,
          85, 255, 255,
         255,  85, 255,
         255, 255,  85,
         255, 255, 255
};


/*
** PcxNextByte(
**   BYTE *byte,              pointer to resultant value
**   int *count,              pointer to repeat count, 1 if no repeat
**   HANDLE fileHandle);      handle of file being read
**
**    Get the next byte from the PCX file, and determine if it is a repeat
**      string or a literal byte.
**
*/


static os_error * PcxNextByte(XBYTE *byte,int *count,buffer * bf)
{
 int value;
 os_error * err;
 *count=1;

 err=bf_getc(bf,&value);
 if(!err)
 {
  /*
   **   Check for a repeat count, indicated by the two high bits being set
   */
   if(0x00c0==(0x00c0 & value))
   {
    *count=0x3f & value;
    err=bf_getc(bf,&value);
   }
   *byte=value;
 }
 return(err);
}









static os_error * ReadPcxFile(buffer * bf,int len)
{
 os_error               * err;
 char                   * palPtr;
 int                      i, j, colors, count, plane;
 XBYTE                    chr, mask, bit, highbit;
 PCXHDR                   pcxHeader;
 PCXPAL256                pcxPal;
 int                      l, imageBytes, bytesPerScanLine;
 char                   * pixels;

 int                      xpix;
 int                      ypix;
 int                      x;
 int                      y;
 int                      xbytes;

 int                      fn;
 imagestr               * loadimage;
 int                    * idata;
 int                      bpp;
 int                      paltype;


 /*
  **      Read in the header information, and verify it
  */

 err=bf_read(bf,&pcxHeader,sizeof(pcxHeader));
 if(err) return(err);


 if(pcxHeader.manufacturer!=10)
 {
  /* error */

 }

 if(pcxHeader.bitsPerPixel*pcxHeader.planes==1)
 {
  colors=2;
 }
 else
 {
  colors=16;
 }

 /*
  **      Set palPtr to point to the palette for this image
  */
 
 palPtr=(char*)pcxHeader.pal1;
 if(pcxHeader.version==5 && len>PCXPALSIZE)
 {
  bf_seek(bf,len-PCXPALSIZE);
  err=bf_read(bf,&pcxPal,PCXPALSIZE);

  if(!err && pcxPal.pal2[0]==12)
  {
   colors=256;
   palPtr=(char*)&pcxPal.pal2[1];
  }
 }

 xpix=1+pcxHeader.maxX-pcxHeader.minX;
 ypix=1+pcxHeader.maxY-pcxHeader.minY;
 bpp=pcxHeader.bitsPerPixel*pcxHeader.planes;
 if(bpp==3) bpp=4;


 pcxHeader.horzDpi=pcxxres;  /* Horizontal dots per inch */
 pcxHeader.vertDpi=pcxyres;

 addframe(loadfile,&fn,0);
 loadfile->frames[fn].fxpix=xpix;
 loadfile->frames[fn].fypix=ypix;
 sprintf(loadfile->frames[fn].name,"PCX%d",fn);
 loadfile->frames[fn].fbpp=bpp;
 loadfile->frames[fn].fxdpi=pcxHeader.horzDpi;
 loadfile->frames[fn].fydpi=pcxHeader.vertDpi;
 loadfile->frames[fn].xim=NULL;


 err=ximnew(loadfile->frames[fn].fxpix,loadfile->frames[fn].fypix,
                          loadfile->frames[fn].fbpp,&loadfile->frames[fn].xim);
 if(err) return(err);

 loadimage=((loadfile->frames[fn].xim)->sim[IM]);
 loadimage->xdpi=loadfile->frames[fn].fxdpi;
 loadimage->ydpi=loadfile->frames[fn].fydpi;


 /*
  **      Fill in intensities for all palette entry colors.  Note the order
  **      change between PCX and DIB.
  */

  loadimage->ipal.ncolours=colors;
  for(i=0;i<colors;++i)
  {
   loadimage->ipal.word[i]=(palPtr[2]<<24)|
                           (palPtr[1]<<16)|
                           (palPtr[0]<<8);
   palPtr+=3;
  }

  /*
   **      Not all PCX files have a valid palette
   */
 
 if(pcxHeader.version==2 || pcxHeader.version>=4)
 {
  for(i=0;i<colors;i++)
  {
   if(loadimage->ipal.word[i]!=0) break;
  }
 }
 else
 {
  i=colors;
 }

 paltype=0;

 if(i==colors)
 {
  if(colors<=16)
  {
   /*
    **      Use default 16 color palette
    */

   paltype=1;

   if(bpp==1)
   {
    loadimage->ipal.word[0]=0;
    loadimage->ipal.word[1]=0xFFFFFF00;
   }
   else
   for(i=0;i<colors;++i)
   {
    loadimage->ipal.word[i]=(DefaultPalette[i*3+2]<<24)|
                            (DefaultPalette[i*3+1]<<16)|
                            (DefaultPalette[i*3+0]<<8);
   }
  }
  else
  {
   /*
    **      Use default Windows Palette
    */

    /* arrgh ! */
   paltype=2;
  }
 }

// getinfo(&loadfile->frames[fn],&pcxHeader,paltype);

 imzero(loadimage);


 /*
  **      Now to do the actual image reading
  */

 bf_seek(bf,sizeof(pcxHeader));
 bytesPerScanLine=pcxHeader.bytesPerLine*pcxHeader.planes;
 imageBytes=bytesPerScanLine*ypix;

 x=0;
 xbytes=0;
 y=0;
 plane=0;

 err=imfind1w(loadimage,y,&idata);
 if(err) return(err);

 pixels=(char*)idata;

 if(bpp==1)
 {
  for(l=0;l<imageBytes;)
  {
   if((err=PcxNextByte(&chr,&count,bf))!=NULL) return(err);

   for(i=0;i<count;i++)
   {
    *pixels++=bitrev[chr];

    xbytes++;
    if(xbytes>=pcxHeader.bytesPerLine)
    {
     y++;

     if(y>=ypix)
     {
      l=imageBytes;
      break;
     }
     else
     {
      xbytes=0;
      err=imfind1w(loadimage,y,&idata);
      if(err) return(err);
      pixels=(char*)idata;
     }
    }
   }
   l+=count;
  }
 }
 else
 if(bpp==2)
 {
  for(l=0;l<imageBytes;)
  {
   if((err=PcxNextByte(&chr,&count,bf))!=NULL) return(err);

   for(i=0;i<count;i++)
   {
    *pixels++=dbitrev[chr];

    xbytes++;
    if(xbytes>=pcxHeader.bytesPerLine)
    {
     y++;

     if(y>=ypix)
     {
      l=imageBytes;
      break;
     }
     else
     {
      xbytes=0;
      err=imfind1w(loadimage,y,&idata);
      if(err) return(err);
      pixels=(char*)idata;
     }
    }
   }
   l+=count;
  }
 }
 else
 if(bpp==4 && pcxHeader.planes>1)
 {
  imzero(loadimage);
  highbit=LowMasks[plane];
  bit=HighMasks[plane];

  for(l=0;l<imageBytes;)
  {
   if((err=PcxNextByte(&chr,&count,bf))!=NULL) return(err);

   for(i=0;i<count;++i)
   {
    for(j=0;j<8;j+=2)
    {
     mask=0x80>>j;

     if(chr & mask)
     {
      *pixels|=highbit;
     }

     mask>>=1;
     if(chr & mask)
     {
      *pixels|=bit;
     }

     pixels++;
    }

    xbytes++;
    if(xbytes>=pcxHeader.bytesPerLine)
    {
     plane++;
     xbytes=0;

     if(plane>=pcxHeader.planes)
     {
      plane=0;
      y++;

      if(y>=ypix)
      {
       l=imageBytes;
       break;
      }
      else
      {
       err=imfind1w(loadimage,y,&idata);
       if(err) return(err);
       pixels=(char*)idata;
       highbit=LowMasks[plane];
       bit=HighMasks[plane];
      }
     }
     else
     {
      pixels=(char*)idata;
      highbit=LowMasks[plane];
      bit=HighMasks[plane];
     }
    }
   } 
   l+=count;
  } 
 }
 else
 if(bpp==4)
 {
  for(l=0;l<imageBytes;)
  {
   if((err=PcxNextByte(&chr,&count,bf))!=NULL) return(err);

   for(i=0;i<count;i++)
   {
    *pixels++=nybrev[chr];

    xbytes++;
    if(xbytes>=pcxHeader.bytesPerLine)
    {
     y++;

     if(y>=ypix)
     {
      l=imageBytes;
      break;
     }
     else
     {
      xbytes=0;
      err=imfind1w(loadimage,y,&idata);
      if(err) return(err);
      pixels=(char*)idata;
     }
    }
   }
   l+=count;
  }
 }
 else
 if(bpp==8)
 {
  for(l=0;l<imageBytes;)
  {
   if((err=PcxNextByte(&chr,&count,bf))!=NULL) return(err);

   for(i=0;i<count;++i)
   {
    *pixels++=chr;

    xbytes++;
    if(xbytes>=pcxHeader.bytesPerLine)
    {
     y++;

     if(y>=ypix)
     {
      l=imageBytes;
      break;
     }
     else
     {
      xbytes=0;
      err=imfind1w(loadimage,y,&idata);
      if(err) return(err);
      pixels=(char*)idata;
     }
    }
   }
   l+=count;
  }
 }
 return(err);
}





os_error * loadpcx(char * name,int type,int userhandle,int xvolatile)
{
 os_error  * err;
 buffer      bf;
 fstat       stat;

 err=fs_stat(name,&stat);

 err=bf_open(name,'r',DEFBUFFSIZE,&bf);
 if(!err)
 {
  err=ReadPcxFile(&bf,stat.length);

  err=bf_close(&bf,err);
 }

 return(err);

 USE(type);
 USE(userhandle);
 USE(xvolatile);
}



/*****************************************************************************/

static int  pcxcount;

static os_error *  pcxputc(buffer * bf,int c)
{
 os_error *  err;
 static int  acc;

 err=NULL;

 if(!pcxcount)
 {
  acc=c;
  pcxcount=1;
 }
 else
 if(c!=acc || pcxcount>=15) 
 {
  if(pcxcount>1 || ((acc & 0xC0)==0xC0))
  {
   bf_putc(bf,0xC0|pcxcount);
  }
  err=bf_putc(bf,acc);

  acc=c;
  pcxcount=1;
 }
 else
 {
  pcxcount++;
 }
 return(err);
}



os_error * savepcx(filestr * file,int fn,char * name,int type)
{
 os_error * err;
 buffer      bf;
 PCXHDR      pcx;

 int         xpix;
 int         ypix;
 int         bpp;
 int         sbpp;
 char      * p;
 int       * idata;
 imagestr  * saveimage;
 int         x;
 int         y;
 int         i;

 int         bit;
 int         highbit;
 int         plane;
 int         ch;
 int         mask;


 saveimage=((file->frames[fn].xim)->sim[IM]);

 xpix=saveimage->xpix;
 ypix=saveimage->ypix;
 bpp=saveimage->bpp;

 pcxcount=0;    /* prime rle coder */

 if(bpp>8)
 {
  err=geterror(EBADBITS);

 }
 else
 {
  err=bf_open(name,'w',DEFBUFFSIZE,&bf);
  if(!err)
  {
   sbpp=(bpp==4)?1:bpp;

   pcx.manufacturer=10;

   if(bpp<8) pcx.version=2;
   else      pcx.version=5;

   pcx.encode=1;
   pcx.bitsPerPixel=sbpp;
   pcx.minX=0;
   pcx.minY=0;
   pcx.maxX=xpix-1;
   pcx.maxY=ypix-1;
   pcx.horzDpi=saveimage->xdpi;
   pcx.vertDpi=saveimage->ydpi;

   memset(&pcx.pal1,0,sizeof(pcx.pal1));

   if(bpp<8)
   {
    p=(char*)&pcx.pal1;

    for(i=0;i<saveimage->ipal.ncolours;i++)
    {
     *p++=saveimage->ipal.word[i]>>8;
     *p++=saveimage->ipal.word[i]>>16;
     *p++=saveimage->ipal.word[i]>>24;
    }
   }

   pcx.junk=0;
   pcx.planes=(bpp==4)?4:1;

   pcx.bytesPerLine=((xpix*sbpp+0x7) & ~0x7)>>3;

   pcx.paltype=1;  /* 1 = Color, 2 = GrayScale */
   pcx.hScreenSize=640;
   pcx.vScreenSize=480;
   memset(&pcx.filler,0,sizeof(pcx.filler));

   err=bf_write(&bf,&pcx,sizeof(pcx));
   if(!err)
   {
    for(y=0;y<ypix;y++)
    {
     err=imfind1r(saveimage,y,&idata);
     if(err) break;
     p=(char*)idata;

     if(bpp==1)
     {
      for(x=0;x<pcx.bytesPerLine;x++)
      {
       pcxputc(&bf,bitrev[*p++]);
      }
     }
     else
     if(bpp==2)
     {
      for(x=0;x<pcx.bytesPerLine;x++)
      {
       pcxputc(&bf,dbitrev[*p++]);
      }
     }
     else
     if(bpp==4 && pcxplanes)
     {
      for(plane=0;plane<4;plane++)
      {
       p=(char*)idata;
       highbit=LowMasks[plane];
       bit=HighMasks[plane];

       for(x=0;x<pcx.bytesPerLine;x++)
       {
        mask=0;

        ch=*p++;
        if(ch & highbit) mask|=0x80;
        if(ch & bit)     mask|=0x40;
        ch=*p++;
        if(ch & highbit) mask|=0x20;
        if(ch & bit)     mask|=0x10;
        ch=*p++;
        if(ch & highbit) mask|=0x08;
        if(ch & bit)     mask|=0x04;
        ch=*p++;
        if(ch & highbit) mask|=0x02;
        if(ch & bit)     mask|=0x01;

        pcxputc(&bf,mask);
       }
      }
     }
     else
     if(bpp==4)
     {
      for(x=0;x<pcx.bytesPerLine;x++)
      {
       pcxputc(&bf,nybrev[*p++]);
      }
     }
     else
     if(bpp==8)
     {
      for(x=0;x<pcx.bytesPerLine;x++)
      {
       pcxputc(&bf,*p++);
      }
     }
    }
    pcxputc(&bf,-1);

    if(bpp==8)
    {
     bf_putc(&bf,12);

     for(i=0;i<saveimage->ipal.ncolours;i++)
     {
      bf_putc(&bf,saveimage->ipal.word[i]>>8);
      bf_putc(&bf,saveimage->ipal.word[i]>>16);
      bf_putc(&bf,saveimage->ipal.word[i]>>24);
     }
    }
   }
   err=bf_closec(&bf,err,name,type);
  }
 }
 return(err);
}


